#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#ifdef UNIX
#include <dlfcn.h>
#endif
#include "tine.h"
#include "mstlib.h"

#ifdef WIN32
# if defined(MSTSIM_EXPORTS) 
# define LIB_EXPORT __declspec(dllexport)
# else
# define LIB_EXPORT __declspec(dllimport)
# endif
#else
# define LIB_EXPORT
#endif

time_t moveto_counts = 0;
int moveto_paused = FALSE;
int moveto_cmd = 0;
#define MOVETO_DURATION 600

int SimulateData(int start,int range)
{
  /* This routine is useful for test servers  */
  return start + rand()%(range);
}
NAME64 mstNames[10];
NAME64 mstNames[10] = {
  "MstSim1", "MstSim2", "MstSim3", "MstSim4", "MstSim5", "MstSim6", "MstSim7", "MstSim8", "MstSim9", "MstSim10", 
};
int mstNum = 10;
float mstPositions[10];
int gNumMotors = 10;

NAME32 stsLst[] = {
  "moving",
  "paused",
  "",
  "",
  "calibrating (step1)",
  "calibrating (step2)",
  "calibrating (step3)",
  "",
  "",
  "",
  "",
  "is calibrating",
  "uncalibrated",
  "",
  "",
  "hardware error",
};

int sizStsLst = sizeof(stsLst)/sizeof(NAME32);

LIB_EXPORT int getStsInfo(int mot,MotorStatus *motsts)
{
  if (motsts == NULL) return invalid_parameter;
  if (mot < 0 || mot > 10) return illegal_device;
  strncpy(motsts->motName,mstNames[mot].name,64);
  motsts->stsBits = mot;
  motsts->stsAxis = 1;
  motsts->stsCoupled = 0;
  motsts->stsClbr = 1;
  motsts->idxCntlLoop = 1;
  motsts->pos = (float)mstPositions[mot];
  motsts->tgt = 0;
  motsts->lag = 0;
  motsts->vel = 0;
  motsts->load = 0;
  motsts->ovrRide = 0;
  return 0;
}
#define LIMIT_UPR 1000
#define LIMIT_LWR 10
void update(void)
{ // simulate some action (this should be a callback from a cdi persistent link)
  static float fval = 10;
  static int dir = 1;
  int i;
  if (dir > 0 && fval > LIMIT_UPR) dir = -1;
  if (dir < 0 && fval < LIMIT_LWR) dir = 1;
  for (i=0; i<10; i++) mstPositions[i] = fval;
  if (!moveto_paused && moveto_counts > 0) 
  {
    fval = fval + (float)(dir * 1.0);
    moveto_counts--;
  }
  if (moveto_counts == 0) moveto_cmd = PROC_STS_IDLE;
}
LIB_EXPORT int getPos(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  if (mot == 0) update();
  *val = (double)mstPositions[mot];
  return 0;
}
LIB_EXPORT int getVel(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  *val =  SimulateData(10,4);
  return 0;
}
LIB_EXPORT int getAcc(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  *val =  SimulateData(5,2);
  return 0;
}

LIB_EXPORT int stepsPerMM(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  *val = 1000;
  return 0;
}

LIB_EXPORT int mstepsPerStep(int mot,int *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  *val = 360;
  return 0;
}

LIB_EXPORT int stepsPerRadian(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  *val = 100/6.2832;
  return 0;
}

LIB_EXPORT int movests(int mot,int *sts)
{
  if (sts == NULL) return invalid_parameter;
  *sts = moveto_cmd;
  return 0;
}
LIB_EXPORT int moveto(int mot,int cmd)
{
  moveto_cmd = cmd;
  switch (cmd)
  {
    case PROC_STS_IDLE:
      moveto_counts = 0;
      moveto_paused = FALSE;
      break;
    case PROC_STS_START:
      moveto_counts = MOVETO_DURATION;
      moveto_paused = FALSE;
      break;
    case PROC_STS_PAUSE:
      moveto_paused = TRUE;
      break;
    case PROC_STS_RESUME:
      moveto_paused = FALSE;
      break;
    case PROC_STS_STOP:
      moveto_counts = 0;
      moveto_paused = FALSE;
      break;
  }
  return 0;
}
typedef struct
{
  int id;
  float gearRatio;
  MotorGroup grp;
  char grpStr[4096];
} SimGrp;

SimGrp grps[10];
int ngrps = 0;

int getSimGrp(void) { return 0; }

LIB_EXPORT int getGrpMembers(int grpId,MotorGroup *grp)
{
  int i;
  for (i=0; i<10; i++)
  {
    if (grps[i].id != grpId) continue;
    memcpy(grp,&grps[i].grp,sizeof(MotorGroup));
    return 0;
  }
  return invalid_index;
}
LIB_EXPORT int setGrpMembers(int grpId,MotorGroup *grp)
{
  int i, n=0, g;
  char *c;

  g = getSimGrp();
  grps[g].id = grpId;
  memcpy(&grps[g].grp,grp,sizeof(MotorGroup));
  c = grps[g].grpStr; *c = 0;
  for (i=0; i<mstNum; i++)
  {
    if (IsMotorInGroup(i,grp))
    {
      if (strlen(c) > 4032) break;
      if (n > 0) strcat(c,",");
      strncat(c,mstNames[i].name,64);
      n++;
    }
  }
  if (ngrps == 0) ngrps++;
  return 0;
}
LIB_EXPORT int setGrpGearRatio(int grpId,double gearRatio)
{
  int i;
  if (gearRatio <= 0) return out_of_range;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    grps[i].gearRatio = (float)gearRatio;
    return 0;
  }
  return invalid_index;
}
LIB_EXPORT int getGrpGearRatio(int grpId,double *gearRatio)
{
  int i;
  if (gearRatio == NULL) return invalid_parameter;
  for (i=0; i<ngrps; i++)
  {
    if (grps[i].id != grpId) continue;
    *gearRatio = (double)grps[i].gearRatio;
    return 0;
  }
  return invalid_index;
}

double tgt = 0;
int itgt = 0;
LIB_EXPORT int setTgtPos(int mot,double val)
{
  if (mot < 0 || mot > 9) return illegal_device;
  tgt = val;
  return 0;
}
LIB_EXPORT int setITgtPos(int mot,int val)
{
  if (mot < 0 || mot > 9) return illegal_device;
  itgt = val;
  return 0;
}
LIB_EXPORT int getTgtPos(int mot,double *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  *val = tgt;
  return 0;
}
LIB_EXPORT int getITgtPos(int mot,int *val)
{
  if (val == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  *val = itgt;
  return 0;
}
LIB_EXPORT int reset(int mot,int *sts)
{
  if (sts == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  *sts = 0;
  return 0;
}
LIB_EXPORT int getSts(int mot,int *sts)
{
  if (sts == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  *sts = 0;
  return 0;
}
LIB_EXPORT int calibr(int mot,int *sts)
{
  if (sts == NULL) return invalid_parameter;
  if (mot < 0 || mot > 9) return illegal_device;
  *sts = 0;
  return 0;
}
LIB_EXPORT int getLastHdwErr(int mot,NAME64 *errstr,int *errcode)
{
  if (errstr == NULL && errcode) return invalid_parameter;
  if (mot < 0 || mot > 10) return illegal_device;
  if (errstr != NULL) 
  { // TODO: fix this
    strncpy(errstr->name,"last error string",64);
  }
  if (errcode != NULL)
  { // TODO: fix this
    *errcode = 0;
  }
  return 0;
}
#define SCANLENGTH 5000
float scanbuffer[SCANLENGTH*2+4];
int scanlength = SCANLENGTH*2 + 1;
LIB_EXPORT int scansts(int mot,DTYPE *dout)
{
  int i, dlen=50;
  float *vals;
  if (mot < 0 || mot > gNumMotors) return illegal_device;
  if (dout == NULL) return invalid_parameter;
  if (dout->dFormat != CF_FLOAT) return illegal_format;
  if (dout->dArrayLength < 50) return dimension_error;
  vals = dout->data.fptr;
  if (dlen > (int)dout->dArrayLength) dlen = (int)dout->dArrayLength;
  dout->dArrayLength = dlen;
  for (i=0; i<25; i++)
  {
    vals[i] = (float)(1+i*0.1);
    vals[i+25] = (float)(5+i);
  }
  return 0;
}
LIB_EXPORT int scanto(int mot,int cmd,DTYPE *din)
{
  din = din;
  return moveto(mot,cmd);
}
int startup(void)
{
  SetMotorList(mstNames,mstNum);
  SetStatusList(stsLst,sizStsLst); 
  SetGetMotorStatusDispatch(getSts);
  SetGetMotorVelocityDispatch(getVel);
  SetGetMotorAccelerationDispatch(getAcc);
  SetGetMotorPositionDispatch(getPos);
  SetGetMotorTargetPositionDispatch(getPos);
  SetGetMotorStepsPerMUDispatch(stepsPerMM);
  SetGetMotorMicroStepsPerStepDispatch(mstepsPerStep);
  SetGetMotorStepsPerRadianDispatch(stepsPerRadian);
  SetGetMoveToStatusDispatch(movests);
  SetGetGroupMoveStatusDispatch(movests);
  SetGetIncrementMoveStatusDispatch(movests);
  SetSetMoveToStatusDispatch(moveto);
  SetSetGroupMoveStatusDispatch(moveto);
  SetSetIncrementMoveStatusDispatch(moveto);
  SetSetMotorTargetPositionDispatch(setTgtPos);
  SetGetMotorTargetPositionDispatch(getTgtPos);
  SetSetMotorTargetMicroPositionDispatch(setITgtPos);
  SetGetMotorTargetMicroPositionDispatch(getITgtPos);
  SetGetGroupMembersDispatch(getGrpMembers);
  SetSetGroupMembersDispatch(setGrpMembers);
  SetCalibrateMotorDispatch(calibr);
  SetResetMotorDispatch(reset);
  SetGetMotorStatusBlockDispatch(getStsInfo);
  SetSetGroupGearRatioDispatch(setGrpGearRatio); // done
  SetGetGroupGearRatioDispatch(getGrpGearRatio); // done
  SetGetLastHardwareErrorDispatch(getLastHdwErr);
  SetGetScanMoveToStatusDispatch(scansts);
  SetSetScanMoveToStatusDispatch(scanto);
  return 0;
}
#ifdef WIN32
BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
      LoadLibrary("cdi32.dll");
      SetInitMotorHardwareDispatch(startup);
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      break;
  }
  return (TRUE);
}
#else
__attribute__((constructor)) void cdiLibMain(void)
{
  dlopen("libcdi.so", RTLD_LAZY);
  SetInitMotorHardwareDispatch(startup);
}
#endif