#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "tine.h"
#if defined(UNIX)
#include "dirent.h"
#elif defined(WIN32)
#include "direct.h"
#endif
#include "arceqm.h"
#include "arctools.h"

void registerStructs(void);
int isAdministrator(char *nam);
extern int useConnectedLinks;
extern char thisExtension[4];

#define EVENTSRVTAG "/SITE/EVENTMGR"
#define DEF_DEV_NAME "#0"

CSVCOLUMN csvTrgChannelColumns[] =
{
  { "CONTEXT","",0,CF_TEXT,32,(BYTE *)OFFSETIN(TrgChannel,ctx),NULL,0 },
  { "SERVER","",0,CF_TEXT,32,(BYTE *)OFFSETIN(TrgChannel,srv),NULL,0 },
  { "DEVICE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgChannel,dev),NULL,0 },
  { "PROPERTY","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgChannel,prp),NULL,0 },
};
CSVDB csvTrgChannelDb =
{
  csvTrgChannelColumns,
  sizeof(csvTrgChannelColumns)/sizeof(csvTrgChannelColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(TrgChannel),
  0
};

CSVCOLUMN csvTrgComponentColumns[] =
{
  { "FILE_NAME","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgComponent,fileName),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgComponent,desc),NULL,0 },
};
CSVDB csvTrgComponentDb =
{
  csvTrgComponentColumns,
  sizeof(csvTrgComponentColumns)/sizeof(csvTrgComponentColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(TrgComponent),
  0
};

TrgViewFile *trgViewFiles = NULL;
int gNumTrgViewFiles = 0;

CSVCOLUMN csvTrgViewFileColumns[] =
{
  { "CONTEXT","",0,CF_TEXT,32,(BYTE *)OFFSETIN(TrgViewFile,ctx),NULL,0 },
  { "TRIGGER","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgViewFile,trg),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgViewFile,desc),NULL,0 },
  { "STRUCTURE_FILE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(TrgViewFile,file),NULL,0 },
};
CSVDB csvTrgViewFileDb =
{
  csvTrgViewFileColumns,
  sizeof(csvTrgViewFileColumns)/sizeof(csvTrgViewFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(TrgViewFile),
  0
};

char cnfgPath[128] = CONFIGS_PATH;
char *getTrcCfgPath(void)
{
  static char path[132] = "";

  if (strlen(path) == 0)
  {
#   if defined(UNIX)
    DIR *dp = NULL;
    if ((dp=opendir(cnfgPath)) == NULL) 
    {
      char tmp[128];
      sprintf(tmp,"%.96s/CONFIGS",ArchiveDirectoryPath);
      if ((dp=opendir(cnfgPath)) != NULL)
      { // legacy directory structure
        strncpy(cnfgPath,tmp,128);
      }
    }
    if (dp != NULL) closedir(dp);
#   endif
    strncpy(path,cnfgPath,128); strcat(path,"/TRC/");
    feclog("configs path: %.132s",path);
  }
  return path;
}
int readTriggerComponentFiles(TrgViewFile *tvf)
{
  int i, nchn, cc=0;
  char fn[64], *path;
  TrgComponent *trg=NULL;
  TrgChannel *chn=NULL;

  path = getTrcCfgPath();
  sprintf(fn,"%.48s",tvf->file);
  if (strstr(fn,".csv") == NULL) strcat(fn,".csv");
  if ((cc=csvReadFile(path,fn,&csvTrgComponentDb,(void **)&trg)) != 0) return cc;
  tvf->components = trg;
  tvf->nComponents = csvTrgComponentDb.siz;
  for (i=0; i<tvf->nComponents; i++)
  { /* go thru the list for the view channels */
    chn = NULL; nchn = 0;
    sprintf(fn,"%.48s",tvf->components[i].fileName);
    if (strstr(fn,".csv") == NULL) strcat(fn,".csv");
    nchn =  (csvReadFile(path,fn,&csvTrgChannelDb,(void **)&chn) == 0) ? csvTrgChannelDb.siz : 0;
    tvf->components[i].channels = chn;
    tvf->components[i].nChannels = nchn;
  }
  return 0;
}

int readViewerFiles(void)
{
  int i, cc=0;
  char *path = getTrcCfgPath();
  /* first read the triggers configuration file */
  if (gNumTrgViewFiles > 0) 
  {
    gNumTrgViewFiles = 0;
    if (trgViewFiles != NULL) free(trgViewFiles);
    trgViewFiles = NULL;
  }
  feclog("looking for TRC viewer configurations in %s",path);
  if ((cc=csvReadFile(path,TRCVIEW_FILE,&csvTrgViewFileDb,(void **)&trgViewFiles)) != 0) ccerr(cc);
  gNumTrgViewFiles = csvTrgViewFileDb.siz;
  feclog("read %d TRC viewer configurations",gNumTrgViewFiles);
  for (i=0; i<gNumTrgViewFiles; i++)
  { /* go thru the list for the view components (ignore file errors here */
    readTriggerComponentFiles(&trgViewFiles[i]);
  }
err:
  return cc;
}
int isViewFileInList(int idx,TrgViewFile *trg)
{
  int i;
  for (i=0; i<idx; i++)
  {
    if (!strncmp(trg[i].file,trg[idx].file,64)) return TRUE;
    if (!strncmp(trg[i].desc,trg[idx].desc,64)) return TRUE;
  }
  return FALSE;
}
int writeTriggerComponentFile(char *config,TrgViewFile *trg,int size)
{
  int i,n,cc = 0;
  char fn[256],tfn[72];
  FILE *fp=NULL;
  TrgComponent *trgcmp=NULL;
  char *path = getTrcCfgPath();

  for (i=0; i<gNumTrgViewFiles; i++) if (!strnicmp(trgViewFiles[i].desc,config,64)) break;
  if (i == gNumTrgViewFiles) return illegal_device;

  strncpy(tfn,trgViewFiles[i].file,64); tfn[64] = 0;
  if (strstr(tfn,".csv") == NULL) strcat(tfn,".csv");
  sprintf(fn,"%.132s%.64s",path,tfn);
  if ((fp=fopen(fn,"wt")) == NULL &&
      (fp=createAndOpen(fn,"wt")) == NULL) ccerr(file_error);
  fprintf(fp,"FileName, Description\n");
  for (n=0; n<size; n++)
  {
    if (isViewFileInList(n,trg)) continue;
    fprintf(fp,"%.64s, %.64s\n",trg[n].file,trg[n].desc);
  }
  fclose(fp); fp = NULL;
  if ((cc=csvReadFile(path,tfn,&csvTrgComponentDb,(void **)&trgcmp)) != 0) ccerr(cc);
  trgViewFiles[i].components = trgcmp;
  trgViewFiles[i].nComponents = csvTrgComponentDb.siz;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int writeTriggerChannelFile(char *desc,TrgChannel *trg,int size)
{
  int i,k,n,nchn=0,cc = 0,found = FALSE;
  char fn[256], cfn[72];
  FILE *fp=NULL;
  TrgChannel *chn=NULL;
  char *path = getTrcCfgPath();

  for (i=0; i<gNumTrgViewFiles; i++)
  {
    for (k=0; k<trgViewFiles[i].nComponents; k++) if (!strnicmp(trgViewFiles[i].components[k].desc,desc,64))
    { /* found it ! */
      strncpy(cfn,trgViewFiles[i].components[k].fileName,64); cfn[64] = 0;
      found = TRUE;
      break;
    }
    if (found) break;
  }
  if (!found) 
  { /* new channel configuration ? */
    return illegal_device;
  }
  if (strstr(cfn,".csv") == NULL) strcat(cfn,".csv");
  sprintf(fn,"%.128s%.64s",path,cfn);
  if ((fp=fopen(fn,"wt")) == NULL &&
      (fp=createAndOpen(fn,"wt")) == NULL) ccerr(file_error);
  fprintf(fp,"Context, Server, Property, Device\n");
  for (n=0; n<size; n++)
  {
    fprintf(fp,"%.32s, %.32s, %.64s, %.64s\n",
      trg[n].ctx,trg[n].srv,trg[n].prp,trg[n].dev);
  }
  fclose(fp); fp = NULL;
  /* now read back what we just wrote ... */
  if (strstr(cfn,".csv") == NULL) strcat(cfn,".csv");
  if ((cc=csvReadFile(path,cfn,&csvTrgChannelDb,(void **)&chn)) == 0) nchn = csvTrgChannelDb.siz;
  trgViewFiles[i].components[k].channels = chn;
  trgViewFiles[i].components[k].nChannels = nchn;
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int writeViewerFiles(TrgViewFile *viewFiles,int nViewFiles)
{
  int cc=0, i;
  char fn[256];
  FILE *fp=NULL;
  char *path = getTrcCfgPath();

  sprintf(fn,"%.132s%s",path,TRCVIEW_FILE);
  
  if ((fp=fopen(fn,"wt")) == NULL &&
      (fp=createAndOpen(fn,"wt")) == NULL) ccerr(file_error);

  fprintf(fp,"Context, Trigger, Description, StructureFile\n");
  for (i=0; i<nViewFiles; i++)
  {
    fprintf(fp,"%.32s, %.64s, %.64s, %.64s\n",
      viewFiles[i].ctx,viewFiles[i].trg,viewFiles[i].desc,viewFiles[i].file);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
int GetTriggerComponents(char *config,TrgChannel *cmp,int cmpBufSize)
{
  int i,n;

  for (i=0; i<gNumTrgViewFiles; i++) if (!strnicmp(trgViewFiles[i].desc,config,64)) break;
  if (i == gNumTrgViewFiles) return -illegal_device;
  for (n=0; n<trgViewFiles[i].nComponents && n<cmpBufSize; n++)
  {
    strncpy(cmp[n].ctx,trgViewFiles[i].ctx,32);
    strncpy(cmp[n].srv,"EVENTS",32);
    strncpy(cmp[n].dev,trgViewFiles[i].components[n].desc,64);
    strncpy(cmp[n].prp,trgViewFiles[i].trg,64);
  }
  return n;
}

int GetTriggerChannels(char *desc,TrgChannel *chn,int chnBufSize)
{
  int i,k,n;

  for (i=0; i<gNumTrgViewFiles; i++) for (k=0; k<trgViewFiles[i].nComponents; k++) 
    if (!strnicmp(trgViewFiles[i].components[k].desc,desc,64))
    { /* found it ! */
      for (n=0; n<chnBufSize && n<trgViewFiles[i].components[k].nChannels; n++)
      {
        strncpy(chn[n].ctx,trgViewFiles[i].components[k].channels[n].ctx,32);
        strncpy(chn[n].srv,trgViewFiles[i].components[k].channels[n].srv,32);
        strncpy(chn[n].dev,trgViewFiles[i].components[k].channels[n].dev,64);
        strncpy(chn[n].prp,trgViewFiles[i].components[k].channels[n].prp,64);
      }
      return n;
    }
  return -illegal_device;
}

AESnotify  aesNotify;
AEScontext aesContext;
ADRB *g_adrb; /* modern archive data request block */
ADRB5 *g_adrb5; /* ancient archive data request block */
ADRBX *g_adrbx; /* old archive data request block */
UINT32 g_eventNumber;

void arcini(void)
{
  static int initialized = 0;

  readViewerFiles();
  if (initialized) return;
  registerStructs();
  initialized = 1;
  return;
}
int GotTriggerEvent = 0;

/* asynchronous callback related routines: */

/* many async requests can be outstanding at the same time
   so we keep a callback table
 */
struct
{
  int id;
  BYTE *buffer;
} callbackTable[100];

int getCallbackSlot(void)
{
  int i;
  for (i=0; i<100; i++) if (callbackTable[i].id == 0) return i;
  return -1;
}
int findCallbackSlot(int id)
{
  int i;
  for (i=0; i<100; i++) if (callbackTable[i].id == id) return i;
  return -1;
}
void (*RecordEventCallback)(int cc) = NULL;
void RcdEventCallback(int id,int cc)
{
  int i = findCallbackSlot(id);
  if (i >= 0)
  {
    callbackTable[i].id = 0; 
  }
  if (RecordEventCallback) (*RecordEventCallback)(cc);
}
void (*SetEventCallback)(UINT32 evnnr,int cc) = NULL;
void SetGlobalEventCallback(int id,int cc)
{
  int i = findCallbackSlot(id);
  if (i >= 0)
  {
    callbackTable[i].id = 0; 
  }
  if (SetEventCallback) (*SetEventCallback)(g_eventNumber,cc);
}

/* for uniqueness, we need to specify a local device id and
   use it as the equipment number.  Although it is not used
   at the server end, it guarantees a unique contract request
   (necessary for asynchronous notification)
*/

#define MAXDEVICES 64 /* this should be plenty */
int ndevices = 0;
char devNameTags[MAXDEVICES][16];

int getDeviceIdFromName(char *devName)
{
  int i;
  for (i=0; i<ndevices && strnicmp(devName,devNameTags[i],16); i++);
  if (i == ndevices)
  {
    if (ndevices == MAXDEVICES) 
    {
      feclog("getDeviceIdFromName: devName %s not found",devName);
      return -illegal_device;
    }
    strncpy(devNameTags[i],devName,16);
    ndevices++;
  }
  return i;
}

PM_EXPORT int RecordEvent(UINT32 eventNumber,UINT32 eventCode,char *devName)
{
  int i,id;
  char devname[48];
  DTYPE din;

  aesContext.version = ARCEQM_VERSION;
  aesContext.eventNumber = eventNumber;
  aesContext.eventCode   = eventCode;
  strncpy(aesContext.deviceName,devName,16);

  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESCON");
  din.dArrayLength = 1;
  din.data.vptr = &aesContext;

  if ((i=getDeviceIdFromName(devName)) < 0) return -i;
  sprintf(devname,"%s/#%d",EVENTSRVTAG,i);
  id = AttachLink(devname,"RECORD",NULL,&din,CA_WRITE,0,RcdEventCallback,CM_SINGLE);
  if (id >= 0 && ((i=getCallbackSlot()) >= 0))
  {
    callbackTable[i].id = id;
  }
  return id < 0 ? -id : 0;
}

int defaultTriggerCallback(UINT32 eventNumber,UINT32 eventCode)
{
  int cc = 0;
  feclog("reacting to event %d code %d",eventNumber,eventCode);
  cc = RecordEvent(eventNumber,eventCode,"NULL");
  return cc;
}
int (*triggerCallback)(UINT32 eventNumber,UINT32 eventCode) = defaultTriggerCallback;
PM_EXPORT void setTriggerCallback(int (*fcn)(UINT32 eventNumber,UINT32 eventCode))
{
  triggerCallback = fcn;
}
PM_EXPORT void setRecordEventCallback(void (*fcn)(int cc))
{
  RecordEventCallback = fcn;
}
PM_EXPORT void setSetEventCallback(void (*fcn)(UINT32 evnnr,int cc))
{
  SetEventCallback = fcn;
}

int (*queryArchiveInfoFunction)(AIB *aib,UINT32 eventNumber,char *deviceServer) = NULL;

PM_EXPORT void setQueryArchiveInfoFunction(int (*fcn)(AIB *aib,UINT32 eventNumber,char *deviceServer))
{
  queryArchiveInfoFunction = fcn;
}
int queryArchiveInfo(AEScontext *aescon,AIB *aib)
{
  int cc = not_implemented;

  if (aib == NULL) return un_allocated;
  if (queryArchiveInfoFunction) cc = (*queryArchiveInfoFunction)(aib,aescon->eventNumber,aescon->deviceName);

  return cc;
}

int (*queryPropertyInfoFunction)(PIB *pib,UINT32 eventNumber,char *deviceServer,char *devProperty) = NULL;
PM_EXPORT void setQueryPropertyInfoFunction(int (*fcn)(PIB *pib,UINT32 eventNumber,char *deviceServer,char *devProperty))
{
  queryPropertyInfoFunction = fcn;
}
int queryPropertyInfo(AEScontext *aescon,char *property,PIB *pib)
{
  int cc = not_implemented;
  if (pib == NULL) return un_allocated;
  if (queryPropertyInfoFunction) 
    cc = (*queryPropertyInfoFunction)(pib,aescon->eventNumber,aescon->deviceName,property);
  return cc;
}
int (*queryPropertyInfoExFunction)(char *fullName,UINT32 evncode,PrpQueryStruct *pqs) = NULL;
PM_EXPORT void setQueryPropertyInfoExFunction(int (*fcn)(char *fullName,UINT32 evncode,PrpQueryStruct *pqs))
{
  queryPropertyInfoExFunction = fcn;
}
int queryPropertyInfoEx(char *fullName,UINT32 evncode,PrpQueryStruct *pqs)
{
  int cc = not_implemented;
  if (pqs == NULL || fullName == NULL) return un_allocated;
  if (queryPropertyInfoExFunction) 
    cc = (*queryPropertyInfoExFunction)(fullName,evncode,pqs);
  return cc;
}

int (*queryNumLocalArchivedEventsFunction)(char *arcctx,UINT32 starttime,UINT32 stoptime,UINT32 eventNumber) = NULL;
PM_EXPORT void setQueryNumLocalArchivedEventsFunction(int (*fcn)(char *arcctx,UINT32 starttime,UINT32 stoptime,UINT32 eventNumber))
{
  queryNumLocalArchivedEventsFunction = fcn;
}
int queryNumLocalArchivedEvents(char *arcctx,UINT32 starttime,UINT32 stoptime,UINT32 eventNumber)
{
  int num = 0;
  if (queryNumLocalArchivedEventsFunction) num = (*queryNumLocalArchivedEventsFunction)(arcctx,starttime,stoptime,eventNumber);
  return num;
}

int (*queryLocalArchivedEventsFunction)(char *arcctx,UINT32 starttime,UINT32 stoptime,UINT32 *evnlst,UINT32 num) = NULL;
PM_EXPORT void setQueryLocalArchivedEventsFunction(int (*fcn)(char *arcctx,UINT32 starttime,UINT32 stoptime,UINT32 *evnlst,UINT32 num))
{
  queryLocalArchivedEventsFunction = fcn;
}
int queryLocalArchivedEvents(char * arcctx,UINT32 starttime,UINT32 stoptime,UINT32 *evnlst,UINT32 num)
{
  int cc = 0;
  if (queryLocalArchivedEventsFunction) cc = (*queryLocalArchivedEventsFunction)(arcctx,starttime,stoptime,evnlst,num);
  return cc;
}

int (*queryNumDeviceArchivedEventsFunction)(char *devsrv,char *device,UINT32 starttime,UINT32 stoptime) = NULL;
PM_EXPORT void setQueryNumDeviceArchivedEventsFunction(int (*fcn)(char *devsrv,char *device,UINT32 starttime,UINT32 stoptime))
{
  queryNumDeviceArchivedEventsFunction = fcn;
}
int queryNumDeviceArchivedEvents(char *devsrv,char *device,UINT32 starttime,UINT32 stoptime)
{
  int cc = 0;
  if (queryNumDeviceArchivedEventsFunction) cc = (*queryNumDeviceArchivedEventsFunction)(devsrv,device,starttime,stoptime);
  return cc;
}

int (*queryDeviceArchivedEventsFunction)(char *arcctx,char *devsrv,char *dev,UINT32 starttime,UINT32 stoptime,AES *aes,UINT32 num,int reverseSort) = NULL;
PM_EXPORT void setQueryDeviceArchivedEventsFunction(int (*fcn)(char *arcctx,char *dev,char *device,UINT32 starttime,UINT32 stoptime,AES *aes,UINT32 num,int reverseSort))
{
  queryDeviceArchivedEventsFunction = fcn;
}
int queryDeviceArchivedEvents(char *arcctx,char *devsrv,char *dev,UINT32 starttime,UINT32 stoptime,AES *aes,UINT32 num,int reverseSort)
{
  int cc = 0;
  if (queryDeviceArchivedEventsFunction) cc = (*queryDeviceArchivedEventsFunction)(arcctx,devsrv,dev,starttime,stoptime,aes,num,reverseSort);
  return cc;
}

int (*queryDeviceArchivedEventListFunction)(char *arcctx,char *srv,char *dev,UINT32 starttime,UINT32 stoptime,UINT32 *evn,UINT32 num,int reverseSort) = NULL;
PM_EXPORT void setQueryDeviceArchivedEventListFunction(int (*fcn)(char *arcctx,char *srv,char *dev,UINT32 starttime,UINT32 stoptime,UINT32 *evn,UINT32 num,int reverseSort))
{
  queryDeviceArchivedEventListFunction = fcn;
}
int queryDeviceArchivedEventList(char *arcctx,char *srv,char *dev,UINT32 starttime,UINT32 stoptime,UINT32 *evn,UINT32 num,int reverseSort)
{
  int cc = 0;
  if (queryDeviceArchivedEventListFunction) cc = (*queryDeviceArchivedEventListFunction)(arcctx,srv,dev,starttime,stoptime,evn,num,reverseSort);
  return cc;
}

int (*queryNumLocalDeviceServersFunction)(void) = NULL;
PM_EXPORT void setQueryNumLocalDeviceServersFunction(int (*fcn)(void))
{
  queryNumLocalDeviceServersFunction = fcn;
}
int queryNumDeviceServers(void)
{
  int num = 0;
  if (queryNumLocalDeviceServersFunction) num = (*queryNumLocalDeviceServersFunction)();
  return num;
}

int (*queryNumLocalDevicePropertiesFunction)(char *devServer,char *devName) = NULL;
PM_EXPORT void setQueryNumLocalDevicePropertiesFunction(int (*fcn)(char *devServer,char *devName))
{
  queryNumLocalDevicePropertiesFunction = fcn;
}
int queryNumDeviceProperties(char *devServer,char *devName)
{
  int num = 0;
  if (queryNumLocalDevicePropertiesFunction) num = (*queryNumLocalDevicePropertiesFunction)(devServer,devName);
  return num;
}

int (*queryNumLocalDevicesFunction)(char *devServer,char *context,int id) = NULL;
PM_EXPORT void setQueryNumLocalDevicesFunction(int (*fcn)(char *devServer,char *context,int id))
{
  queryNumLocalDevicesFunction = fcn;
}
int queryNumDevices(char *devServer,char *context,int id)
{
  int num = 0;
  if (queryNumLocalDevicesFunction) num = (*queryNumLocalDevicesFunction)(devServer,context,id);
  return num;
}

int (*queryLocalDeviceServerNamesFunction)(char *arcctx,NAME32 *names,UINT32 start,UINT32 stop,int event,int num) = NULL;
PM_EXPORT void setQueryLocalDeviceServerNamesFunction(int (*fcn)(char *arcctx,NAME32 *names,UINT32 start,UINT32 stop,int event,int num))
{
  queryLocalDeviceServerNamesFunction= fcn;
}
int (*queryLocalDeviceServerNamesExFunction)(char *arcctx,NAME64 *names,UINT32 start,UINT32 stop,int event,int num) = NULL;
PM_EXPORT void setQueryLocalDeviceServerNamesExFunction(int (*fcn)(char *arcctx,NAME64 *names,UINT32 start,UINT32 stop,int event,int num))
{
  queryLocalDeviceServerNamesExFunction= fcn;
}
int queryDeviceServerNames(char *arcctx,NAME32 *names,UINT32 start,UINT32 stop,int event,int num)
{
  int cc = 0;
  if (queryLocalDeviceServerNamesFunction) cc = (*queryLocalDeviceServerNamesFunction)(arcctx,names,start,stop,event,num);
  return cc;
}
int queryDeviceServerNamesEx(char *arcctx,NAME64 *names,UINT32 start,UINT32 stop,int event,int num)
{
  int cc = 0;
  if (queryLocalDeviceServerNamesExFunction) cc = (*queryLocalDeviceServerNamesExFunction)(arcctx,names,start,stop,event,num);
  return cc;
}

int (*queryLocalDevicesFunction)(NAME64 *names,char *devServer,char *context,int id,int num) = NULL;
PM_EXPORT void setQueryLocalDeviceNamesFunction(int (*fcn)(NAME64 *names,char *devServer,char *context,int id,int num))
{
  queryLocalDevicesFunction= fcn;
}
int queryDevices(NAME64 *names,char *devServer,char *context,int id,int num)
{
  int cc = 0;
  if (queryLocalDevicesFunction) cc = (*queryLocalDevicesFunction)(names,devServer,context,id,num);
  return cc;
}

int (*queryLocalDevicePropertiesFunction)(NAME64 *names,char *devServer,char *devName,int num) = NULL;
PM_EXPORT void setQueryLocalDevicePropertiesFunction(int (*fcn)(NAME64 *names,char *devServer,char *devName,int num))
{
  queryLocalDevicePropertiesFunction = fcn;
}
int queryDeviceProperties(NAME64 *names,char *devServer,char *devName,int num)
{
  int cc = 0;
  if (queryLocalDevicePropertiesFunction) cc = (*queryLocalDevicePropertiesFunction)(names,devServer,devName,num);
  return cc;
}

int (*queryLocalTriggersFunction)(NAME64 *names,char *devServer,char *prpName,int num) = NULL;
PM_EXPORT void setQueryLocalTriggersFunction(int (*fcn)(NAME64 *names,char *devServer,char *prpName,int num))
{
  queryLocalTriggersFunction = fcn;
}
int queryTriggerNames(NAME64 *names,char *devServer,char *prpName,int num)
{
  int cc = 0;
  if (queryLocalTriggersFunction) cc = (*queryLocalTriggersFunction)(names,devServer,prpName,num);
  return cc;
}
void arctsk(void)
{
  if (GotTriggerEvent)
  {
    if (triggerCallback) (*triggerCallback)(aesNotify.eventNumber,aesNotify.eventCode);
    GotTriggerEvent = 0;
  }
}
PM_EXPORT int SetGlobalEventComment(UINT32 eventCode,char *cmt,char *mode)
{
  AESnotify aesn;
  char devname[48];
  short modeNo = strchr(mode,'a') ? 1 : 0;
  DTYPE din;

  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESNOT");
  din.dArrayLength = 1;
  din.data.vptr = &aesn;
  aesn.version = ARCEQM_VERSION;
  aesn.eventCode = eventCode;
  strncpy(aesn.comment[0],cmt,sizeof(aesn.comment));
  sprintf(devname,"%s/#%d",EVENTSRVTAG,modeNo);
  return ExecLink(devname,"SETCOMMENT",NULL,&din,CA_WRITE);
}
PM_EXPORT int SetGlobalEvent(UINT32 eventCode,char *cmt)
{
  int i,id;
  static AESnotify aesn;
  DTYPE din,dout;

  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"AESNOT");
  din.dArrayLength = 1;
  din.data.vptr = &aesn;

  dout.dFormat = CF_LONG;
  dout.dArrayLength = 1;
  g_eventNumber = time(NULL); /* set it to now in case the call fails! */
  dout.data.ulptr = &g_eventNumber;

  aesn.version = ARCEQM_VERSION;
  aesn.eventCode = eventCode;  
  strncpy(aesn.comment[0],cmt,64);
  aesn.comment[1][0] = 0; aesn.comment[2][0] = 0; aesn.comment[3][0] = 0;

  id = AttachLink(EVENTSRVTAG,"SET",&dout,&din,CA_WRITE|CA_READ,0,SetGlobalEventCallback,CM_SINGLE);
  if (id >= 0 && ((i=getCallbackSlot()) >= 0))
  {
    callbackTable[i].id = id;
  }
  return id < 0 ? -id : 0;
}

int _AnnotateArchiveEventFunction(UINT32 eventNumber,char *dsrv,char *cmt) { return not_implemented; }
int (*AnnotateArchiveEventFunction)(UINT32 eventNumber,char *dsrv,char *cmt) = _AnnotateArchiveEventFunction;
PM_EXPORT void setAnnotateArchiveEventFunction(int (*fcn)(UINT32 eventNumber,char *dsrv,char *cmt))
{
  if (fcn != NULL) AnnotateArchiveEventFunction = fcn;
}

int _RemoveArchiveEventsFunction(UINT32 eventStart,UINT32 eventLast,char *dsrv) { return not_implemented; }
int (*RemoveArchiveEventsFunction)(UINT32 eventFirst,UINT32 eventLast,char *dsrv) = _RemoveArchiveEventsFunction;
PM_EXPORT void setRemoveArchiveEventsFunction(int (*fcn)(UINT32 eventFirst,UINT32 eventLast,char *dsrv))
{
  if (fcn != NULL) RemoveArchiveEventsFunction = fcn;
}
int _RemoveArchiveEventFunction(UINT32 eventNumber,char *dsrv) { return not_implemented; }
int (*RemoveArchiveEventFunction)(UINT32 eventNumber,char *dsrv) = _RemoveArchiveEventFunction;
PM_EXPORT void setRemoveArchiveEventFunction(int (*fcn)(UINT32 eventNumber,char *dsrv))
{
  if (fcn != NULL) RemoveArchiveEventFunction = fcn;
}

/* ancient style: */
int (*GetArchiveDataFunction)(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC5 *devspec) = NULL;
PM_EXPORT void setGetArchiveDataFunction(int (*fcn)(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC5 *devspec))
{
  GetArchiveDataFunction = fcn;
}
int _GetArchiveData(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC5 *devspec)
{
   if (GetArchiveDataFunction == NULL) return not_implemented;
   if (tineDebug) dbglog("GetData : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return (*GetArchiveDataFunction)(d,evnNumber,starttime,stoptime,evnContext,devspec);
}

/* old style: */
int (*GetArchiveDataFunctionEx)(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPECX *devspec) = NULL;
PM_EXPORT void setGetArchiveDataFunctionEx(int (*fcn)(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPECX *devspec))
{
  GetArchiveDataFunctionEx = fcn;
}
int _GetArchiveDataEx(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPECX *devspec)
{
   if (GetArchiveDataFunctionEx == NULL) return not_implemented;
   if (tineDebug) dbglog("GetData : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return (*GetArchiveDataFunctionEx)(d,evnNumber,starttime,stoptime,evnContext,devspec);
}
/* new style: */
int (*GetDataFunction)(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devspec) = NULL;
PM_EXPORT void setGetDataFunction(int (*fcn)(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devSpec))
{
  GetDataFunction = fcn;
}
int _GetData(DTYPE *d,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devspec)
{
   if (GetDataFunction == NULL) return not_implemented;
   if (tineDebug) dbglog("GetData : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return (*GetDataFunction)(d,evnNumber,starttime,stoptime,evnContext,devspec);
}

/* ancient style: */
int (*GetArchiveHeaderFunction)(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC5 *devspec) = NULL;
PM_EXPORT void setGetArchiveHeaderFunction(int (*fcn)(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC5 *devspec))
{
  GetArchiveHeaderFunction = fcn;
}
int _GetArchiveHeader(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC5 *devspec)
{
   if (GetArchiveHeaderFunction == NULL) return not_implemented;
   if (tineDebug) dbglog("GetHeader : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return (*GetArchiveHeaderFunction)((DHS *)dhs,evnNumber,starttime,stoptime,evnContext,devspec);
}

/* old style: */
int (*GetArchiveHeaderFunctionEx)(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPECX *devspec) = NULL;
PM_EXPORT void setGetArchiveHeaderFunctionEx(int (*fcn)(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPECX *devspec))
{
  GetArchiveHeaderFunctionEx = fcn;
}
int _GetArchiveHeaderEx(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPECX *devspec)
{
   if (GetArchiveHeaderFunctionEx == NULL) return not_implemented;
   if (tineDebug) dbglog("GetHeader : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return (*GetArchiveHeaderFunctionEx)((DHS *)dhs,evnNumber,starttime,stoptime,evnContext,devspec);
}

/* new style: (but old output header) */
int (*GetHeaderFunction)(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devspec) = NULL;
PM_EXPORT void setGetHeaderFunction(int (*fcn)(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devspec))
{
  GetHeaderFunction = fcn;
}
int _GetHeader(DHS *dhs,UINT32 evnNumber,UINT32 starttime,UINT32 stoptime,char *evnContext,DEVSPEC *devspec)
{
   if (GetHeaderFunction == NULL) return not_implemented;
   if (tineDebug) dbglog("GetHeader : event %d (%d to %d)",evnNumber,starttime,stoptime);
   return (*GetHeaderFunction)((DHS *)dhs,evnNumber,starttime,stoptime,evnContext,devspec);
}

/* newest style: */
int (*GetHdrFcn)(DTYPE *dout,UINT32 evnNumber,char *evnContext,DEVSPEC *devspec) = NULL;
PM_EXPORT void setGetHeaderDispatch(int (*fcn)(DTYPE *dout,UINT32 evnNumber,char *evnContext,DEVSPEC *devspec))
{
  GetHdrFcn = fcn;
}
int getHdr(DTYPE *dout,UINT32 evnNumber,char *evnContext,DEVSPEC *devspec)
{
   if (GetHdrFcn == NULL) return not_implemented;
   if (tineDebug) dbglog("GetHeader : event %d (%s: %s %s)",
     evnNumber,evnContext,devspec->deviceServerName,devspec->deviceProperty);
   return (*GetHdrFcn)(dout,evnNumber,evnContext,devspec);
}

int (*CloseFileFcn)(UINT32 evnNumber,char *evnTrigger) = NULL;
PM_EXPORT void setCloseArchiveFileDispatch(int (*fcn)(UINT32 evnNumber,char *evnTrigger))
{
  CloseFileFcn = fcn;
}

int closeArcFile(UINT32 evnNumber,char *evnTrigger)
{
   if (CloseFileFcn == NULL) return not_implemented;
   if (tineDebug) dbglog("CloseArchiveFile : event %d trigger %s",evnNumber,evnTrigger);
   return (*CloseFileFcn)(evnNumber,evnTrigger);
}

UINT32 archiveState = 0;
UINT32 archiveWindow = 30;
enum {
  TriggerNormal = 0,
  TriggerNext,
  TriggerWindow,
  TriggerRefuse
} archiveStates;
int isTriggerAllowed(void)
{
  static time_t lasttrigger = 0;
  time_t t = time(NULL);
  
  switch (archiveState)
  {
    case TriggerNormal: 
      lasttrigger = t;
      return TRUE;
    case TriggerNext:
      lasttrigger = t;
      archiveState = TriggerRefuse;
      return TRUE;
    case TriggerWindow:
      if (t-lasttrigger > (time_t)archiveWindow)
      {
        lasttrigger = t;
        return TRUE;
      }
      return FALSE;
    case TriggerRefuse:
      return FALSE;
  }
  return FALSE;
}
PM_EXPORT int arceqm(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  char caller[USERNAME_SIZE+8];
  time_t eventTime;
  int eventNumber,nevn;
  int i,k,p,n,nlen,cc,reverseSort=FALSE;
  static AES aes[20];
  NAME32 *n32;
  NAME64 *n64;
  UINT32 starttime,stoptime,evncode;
  char devServerName[128];
  char device[DEVICE_NAME_SIZE];
  DEVSPEC devsp;
  char *c,*ctx,*srv,*dev,*prp,*trg;

  memset(devServerName,0,128);
  memset(device,0,DEVICE_NAME_SIZE);
  strncpy(caller,GetCaller(ARCEQM_TAG),USERNAME_SIZE);
  if (!stricmp(devProperty,"TRIGGER"))  
  { /* respond to external trigger request (defunct and deprecated) */
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (!isTriggerAllowed()) return not_accepted;
    if ( din->dFormat != CF_STRUCT ||
         strcmp(din->dTag,"AESNOT") ) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    memcpy(&aesNotify,din->data.vptr,sizeof(AESnotify));
    if (aesNotify.eventCode == EVENT_IGNORE) return 0;
    if (tineDebug) dbglog("EVNEQM: received TRIGGER %d from %.16s",aesNotify.eventNumber,caller);
    /* set trigger flag so that the call returns quickly */
    GotTriggerEvent = 1;
    /* handle the trigger asynchronously */
    return 0;
  }
  if (!stricmp(devProperty,"ARCHINFO") || !stricmp(devProperty,"ArchiveInformation"))  
  { /* archive info data */
    if (access & CA_WRITE) return illegal_read_write;
    if ( din->dFormat != CF_STRUCT ||
         strcmp(din->dTag,"AESCON") ) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    if ( dout->dFormat != CF_STRUCT ||
         strcmp(dout->dTag,"AIB") ) return illegal_format;
    if (dout->dArrayLength != 1) return dimension_error;
    return queryArchiveInfo((AEScontext *)din->data.vptr,(AIB *)dout->data.vptr);
  }
  if (!stricmp(devProperty,"ARCHSTATE") || !stricmp(devProperty,"ArchiveTriggerStatus"))
  {
    if (access & CA_WRITE)
    {
      if ((cc=getValuesAsLong(din,(SINT32 *)&archiveState,1)) != 0) return cc;
    } 
    if (dout->dArrayLength > 0) cc = putValuesFromUnsignedLong(dout,&archiveState,1);
    return cc;
  }
  if (!stricmp(devProperty,"ARCHWINDOW") || !stricmp(devProperty,"ArchiveTriggerWindow"))
  {
    if (access & CA_WRITE)
    {
      if ((cc=getValuesAsLong(din,(SINT32 *)&archiveWindow,1)) != 0) return cc;
    }
    if (dout->dArrayLength > 0) cc = putValuesFromUnsignedLong(dout,&archiveWindow,1);
    return cc;
  }
  if (!stricmp(devProperty,"PROPINFO"))  
  { /* property info data */
    if (access & CA_WRITE) return illegal_read_write;
    if (devProperty[8] != '.') return illegal_delimiter;
    if ( din->dFormat != CF_STRUCT ||
         strcmp(din->dTag,"AESCON") ) return illegal_format;
    if (din->dArrayLength != 1) return dimension_error;
    if ( dout->dFormat != CF_STRUCT ||
         strcmp(dout->dTag,"PIB") ) return illegal_format;
    if (dout->dArrayLength != 1) return dimension_error;

    return queryPropertyInfo((AEScontext *)din->data.vptr,&devProperty[9],(PIB *)dout->data.vptr);
  }
  if (!stricmp(devProperty,"PropertyInformation"))  
  { /* property info data */
    if (access & CA_WRITE) return illegal_read_write;
    evncode = 0;
    if (din->dArrayLength > 0)
    {
      if ((cc=GetValuesAsLong(din,(SINT32 *)&evncode,1)) != 0) return cc;
    }
    if (dout->dFormat != CF_STRUCT || strcmp(dout->dTag,"PRPQSr4") ) return illegal_format;
    if (dout->dArrayLength != 1) return dimension_error;
    return queryPropertyInfoEx(devName,evncode,(PrpQueryStruct *)dout->data.vptr);
  }
  if (!stricmp(devProperty,"CloseArchive"))
  {
    if (access & CA_WRITE) return illegal_read_write;
    if (din->dArrayLength  < 1) return dimension_error;
    if (din->dFormat != CF_INT32) return illegal_format;
    evncode = din->data.lptr[0];
    return closeArcFile(evncode,devName);
  }
  if (!stricmp(devProperty,"ARCHDATA") || !stricmp(devProperty,"StoredData"))  
  { /* respond to request for archive data */
    if (access & CA_WRITE) return illegal_read_write;
    if (din->dArrayLength  < 1) return dimension_error;
    if (dout->dArrayLength < 1) return dimension_error;
    switch (din->dFormat)
    {
      case CF_LONG:
        stoptime = time(NULL);
        starttime = stoptime - 3600 * 24 * 3;
        switch (din->dArrayLength)
        {
          case 3:
            stoptime = din->data.lptr[2];
          case 2:
            starttime = din->data.lptr[1];
          case 1:
            evncode = din->data.lptr[0];
            break;
          default:
            return dimension_error;
        }
        /* devName must be of the form: <trigger>/<context>/<server>/<device>[property] */
        if ((c=strchr(devName,'/')) == NULL) return argument_list_error;
        if ((c=strchr(devName,':')) != NULL) 
        { /* or of the form: <trigger>:/<context>/<server>/<device>[property] */
          *c++ = 0; 
          if (*c != '/') return argument_list_error;
        }
        *c++ = 0; ctx = c;
        if ((c=strchr(ctx,'/')) == NULL) return argument_list_error;
        *c++ = 0; srv = c;
        if ((c=strchr(srv,'/')) == NULL) return argument_list_error;
        *c++ = 0; dev = c;
        if ((c=strchr(dev,'[')) == NULL) return argument_list_error;
        *c++ = 0; prp = c;
        if ((c=strchr(prp,']')) == NULL) return argument_list_error;
        *c = 0;
        strncpy(devsp.deviceContext,ctx,CONTEXT_NAME_SIZE);
        strncpy(devsp.deviceServerName,srv,EXPORT_NAME_SIZE);
        strncpy(devsp.deviceProperty,prp,PROPERTY_NAME_SIZE);
        strncpy(devsp.deviceName,dev,DEVICE_NAME_SIZE);
        memset(devsp.dataTag,0,TAG_NAME_SIZE);
        devsp.dataFormat = dout->dFormat;
        devsp.dataIndex = 0;
        devsp.dataDimension = dout->dArrayLength;
        n = _GetData(dout,evncode,starttime,stoptime,devName,&devsp);
        if (n < 0) return -n;
        dout->dArrayLength = n;
        return 0;
      case CF_STRUCT:
        if (!strcmp(din->dTag,"ADRBr4"))
        {
          g_adrb = (ADRB *)din->data.vptr;
          g_adrb->devSpecifier.dataDimension = dout->dArrayLength;
          g_adrb->devSpecifier.dataFormat = dout->dFormat;
          strncpy(g_adrb->devSpecifier.dataTag,dout->dTag,TAG_NAME_SIZE);
          /* use devName as event context */
          n = _GetData(dout,g_adrb->eventNumber,g_adrb->starttime,g_adrb->stoptime,devName,&g_adrb->devSpecifier);
          if (n < 0) return -n;
          dout->dArrayLength = n;
          return 0;
        }
        if (!strcmp(din->dTag,"ADRB"))
        {
          g_adrb5 = (ADRB5 *)din->data.vptr;
          g_adrb5->devSpecifier.dataDimension = dout->dArrayLength;
          g_adrb5->devSpecifier.dataFormat = dout->dFormat;
          /* use devName as event context */
          n = _GetArchiveData(dout,g_adrb5->eventNumber,g_adrb5->starttime,g_adrb5->stoptime,devName,&g_adrb5->devSpecifier);
          if (n < 0) return -n;
          dout->dArrayLength = n;
          return 0;
        }
        if (!strcmp(din->dTag,"ADRBX"))
        {
          g_adrbx = (ADRBX *)din->data.vptr;
          g_adrbx->devSpecifier.dataDimension = dout->dArrayLength;
          g_adrbx->devSpecifier.dataFormat = dout->dFormat;
          /* use devName as event context */
          n = _GetArchiveDataEx(dout,g_adrbx->eventNumber,g_adrbx->starttime,g_adrbx->stoptime,devName,&g_adrbx->devSpecifier);
          if (n < 0) return -n;
          dout->dArrayLength = n;
          return 0;
        }
      default:
        return illegal_format;
    }
  }

  if (!stricmp(devProperty,"ARCHDATA.NAM") || !stricmp(devProperty,"StoredData.NAM") ||
      !stricmp(devProperty,"ARCHHEADER.NAM") || !stricmp(devProperty,"DataHeader.NAM") ||
      !stricmp(devProperty,"CloseArchive.NAM") ||
      !stricmp(devProperty,"ANNOTATE.NAM") || !stricmp(devProperty,"REMOVE.NAM"))
  {
    if (dout->dArrayLength > (UINT32)arbTblSiz) dout->dArrayLength = (UINT32)arbTblSiz;
    return PutValuesFromAny(dout,arbTbl[0].trg,CF_NAME64,1,sizeof(ARB),arbTblSiz,0);
  }
  if (!stricmp(devProperty,"ARCHHEADER") || !stricmp(devProperty,"DataHeader"))  
  { /* respond to request for archive data header */
    /* n.b. 'starttime' and 'stoptime' are fairly superfluous below: the event Nr says it all */
    if (access & CA_WRITE) 
    { /* can update static header information only */
      if (din->dArrayLength < 1) return dimension_error;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,"PHSr4")) return invalid_structure_tag;
      return dumpPHS((PHS *)din->data.vptr,din->dArrayLength);
    }
    if (dout->dFormat != CF_STRUCT) return illegal_format;
    if (dout->dArrayLength < 1) return dimension_error;
    if (din->dArrayLength < 1) return dimension_error;
    switch (din->dFormat)
    {
      case CF_LONG:
        stoptime = time(NULL);
        starttime = stoptime - 3600 * 24 * 3;
        switch (din->dArrayLength)
        {
          case 3:
            stoptime = din->data.lptr[2];
          case 2:
            starttime = din->data.lptr[1];
          case 1:
            evncode = din->data.lptr[0];
            break;
          default:
            return dimension_error;
        }
        /* devName must be of the form: <event context>/<context>/<server>/<device>[property] */
        if ((c=strchr(devName,'/')) == NULL) return argument_list_error;
        *c++ = 0; ctx = c;
        if ((c=strchr(ctx,'/')) == NULL) return argument_list_error;
        *c++ = 0; srv = c;
        if ((c=strchr(srv,'/')) == NULL) return argument_list_error;
        *c++ = 0; dev = c;
        if ((c=strchr(dev,'[')) == NULL) return argument_list_error;
        *c++ = 0; prp = c;
        if ((c=strchr(prp,']')) == NULL) return argument_list_error;
        *c = 0;
        strncpy(devsp.deviceContext,ctx,CONTEXT_NAME_SIZE);
        strncpy(devsp.deviceServerName,srv,EXPORT_NAME_SIZE);
        strncpy(devsp.deviceProperty,prp,PROPERTY_NAME_SIZE);
        strncpy(devsp.deviceName,dev,DEVICE_NAME_SIZE);
        memset(devsp.dataTag,0,TAG_NAME_SIZE);
        devsp.dataFormat = dout->dFormat;
        devsp.dataIndex = 0;
        devsp.dataDimension = dout->dArrayLength;
        /* devName now gives the event context */
        break;
      case CF_STRUCT:
        if (!strcmp(din->dTag,"ADRBr4"))
        {
          g_adrb = (ADRB *)din->data.vptr;
          memcpy(&devsp,&g_adrb->devSpecifier,sizeof(DEVSPEC));
          evncode = g_adrb->eventNumber;
          break;
        }
        if (!strcmp(din->dTag,"ADRB"))
        {
          g_adrb5 = (ADRB5 *)din->data.vptr;
          strncpy(devsp.deviceContext,gDeviceContext,CONTEXT_NAME_SIZE);
          strncpy(devsp.deviceServerName,g_adrb5->devSpecifier.deviceServerName,16);
          strncpy(devsp.deviceProperty,g_adrb5->devSpecifier.deviceProperty,16);
          strncpy(devsp.deviceName,g_adrb5->devSpecifier.deviceName,16);
          memset(devsp.dataTag,0,TAG_NAME_SIZE);
          devsp.dataFormat = dout->dFormat;
          devsp.dataIndex = 0;
          devsp.dataDimension = dout->dArrayLength;
          evncode = g_adrb5->eventNumber;
          break;
        }
        if (!strcmp(din->dTag,"ADRBX"))
        {
          g_adrbx = (ADRBX *)din->data.vptr;
          strncpy(devsp.deviceContext,g_adrbx->devSpecifier.deviceContext,CONTEXT_NAME_SHORTSIZE);
          strncpy(devsp.deviceServerName,g_adrbx->devSpecifier.deviceServerName,EXPORT_NAME_SHORTSIZE);
          strncpy(devsp.deviceProperty,g_adrbx->devSpecifier.deviceProperty,PROPERTY_NAME_SHORTSIZE);
          strncpy(devsp.deviceName,g_adrbx->devSpecifier.deviceName,32);
          memset(devsp.dataTag,0,TAG_NAME_SIZE);
          devsp.dataFormat = dout->dFormat;
          devsp.dataIndex = 0;
          devsp.dataDimension = dout->dArrayLength;
          evncode = g_adrbx->eventNumber;
          break;
        }
      default:
        return illegal_format;
    }
    return getHdr(dout,evncode,devName,&devsp);
  }
  if (!stricmp(devProperty,"NumEvents") || !stricmp(devProperty,"NEvents"))
  {
    if (din->dFormat != CF_LONG) return illegal_format;
    if (dout->dFormat != CF_LONG) return illegal_format;
    evncode = 0;
    switch (din->dArrayLength)
    {
      case 3: evncode   = din->data.lptr[2];
      case 2: starttime = din->data.lptr[0];
              stoptime  = din->data.lptr[1];
              break;
      default: return illegal_read_write;
    }
    *dout->data.lptr = queryNumLocalArchivedEvents(devName,starttime,stoptime,evncode);
    return 0;
  }
  if (!stricmp(devProperty,"Events"))
  {
    stoptime = time(NULL);
    starttime = stoptime - 3600 * 24 * 3;
    if (din->dArrayLength > 0)
    {
      if (din->dFormat != CF_LONG) return illegal_format;
      starttime = din->data.lptr[0];
      if (din->dArrayLength > 1) stoptime  = din->data.lptr[1];
    }
    if (dout->dFormat != CF_LONG) return illegal_format;
    /* devName either trigger (1st choice) or server name */
    n = queryLocalArchivedEvents(devName,starttime,stoptime,dout->data.ulptr,dout->dArrayLength);
    if (n < 0) return -n;
    dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"StoredContexts") || 
      !stricmp(devProperty,"Events.NAM") || 
      !stricmp(devProperty,"NEvents.NAM") ||
      !stricmp(devProperty,"NumEvents.NAM") ||
      !stricmp(devProperty,"StoredServers.NAM") )
  {
    char prp[DEVICE_NAME_SIZE] = { "" };
    if (strlen(devName) > 0 && devName[0] != '#') strncpy(devServerName,devName,DEVICE_NAME_SIZE);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength != 1) return illegal_read_write;
      switch (din->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
          nlen = fmtsizeof(din->dFormat);
          break;
        default:
          return illegal_format;
      }
      strncpy(prp,din->data.cptr,nlen);
    }
    switch (dout->dFormat)
    {
      case CF_NAME64:
        n64 = (NAME64 *)dout->data.vptr;
        break;
      case CF_NAME32:
      case CF_NAME16:
        n64 = (NAME64 *)qryMemPool;
        break;
      default:
        return illegal_format;
    }
    n = queryTriggerNames(n64,devServerName,prp,dout->dArrayLength);
    if (n < 0) return -n;
    switch (dout->dFormat)
    {
      case CF_NAME32:
        N1toN2((NAME32 *)dout->data.vptr,n64,dout->dArrayLength,32);
        break;
      case CF_NAME16:
        N1toN2((NAME16 *)dout->data.vptr,n64,dout->dArrayLength,16);
        break;
    }
    dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"NDSERVERS") || !stricmp(devProperty,"NumStoredServers"))
  {
    if (dout->dArrayLength != 1) return illegal_read_write;
    switch (dout->dFormat)
    {
      case CF_INTEGER: *dout->data.sptr = queryNumDeviceServers(); break;
      case CF_LONG: *dout->data.lptr = queryNumDeviceServers(); break;
      default: return illegal_format;
    }
    return 0;
  }
  if (!stricmp(devProperty,"DSERVERS") || !stricmp(devProperty,"StoredServers"))
  { // devName = trigger name
    starttime = 0; stoptime = time(NULL);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength == 1 && dout->dFormat == CF_NAME32)
      {
        n = dout->dArrayLength;
        n32 = (NAME32 *)dout->data.vptr;
        cc = getArchiveServers(din->data.lptr[0],devName,NULL,n32,&n);
        dout->dArrayLength = n;
        return cc;
      }
      if (din->dArrayLength != 2) return illegal_read_write;
      if (din->dFormat != CF_LONG) return illegal_format;
      starttime = din->data.lptr[0];
      stoptime = din->data.lptr[1];
    }
    switch (dout->dFormat)
    {
      case CF_NAME64:
        n = queryDeviceServerNamesEx(devName,(NAME64 *)dout->data.vptr,starttime,stoptime,0,(int)dout->dArrayLength);
        if (n >= 0) dout->dArrayLength = n;
        return 0;
      case CF_NAME32:
        n32 = (NAME32 *)dout->data.vptr;
        break;
      case CF_NAME16:
        n32 = (NAME32 *)qryMemPool;
        break;
      default:
        return illegal_format;
    }
    n = queryDeviceServerNames(devName,n32,starttime,stoptime,0,(int)dout->dArrayLength);
    if (dout->dFormat == CF_NAME16 && n > 0) 
      N1toN2((NAME16 *)dout->data.vptr,n32,dout->dArrayLength,16);
    if (n >= 0) dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"NAPROPS") || !stricmp(devProperty,"NumStoredProperties"))
  {
    char trgstr[DEVICE_NAME_SIZE] = { "" };
    if (strlen(devName) == 0) return illegal_device_number;
    strncpy(devServerName,devName,DEVICE_NAME_SIZE);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength != 1) return illegal_read_write;
      switch (din->dFormat)
      {
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
          nlen = fmtsizeof(din->dFormat);
          break;
        default:
          return illegal_format;
      }
      strncpy(trgstr,din->data.cptr,nlen);
    }
    if (dout->dArrayLength != 1) return illegal_read_write;
    switch (dout->dFormat)
    {
      case CF_INTEGER: *dout->data.sptr = queryNumDeviceProperties(devServerName,trgstr); break;
      case CF_LONG: *dout->data.lptr = queryNumDeviceProperties(devServerName,trgstr); break;
      default: return illegal_format;
    }
    return 0;
  }

  if (!stricmp(devProperty,"APROPS") || !stricmp(devProperty,"StoredProperties"))
  {
    char devn[DEVICE_NAME_SIZE] = { "" };
    if (strlen(devName) == 0) return illegal_device_number;
    strncpy(devServerName,devName,128);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength != 1) return illegal_read_write;
      switch (din->dFormat)
      {
        case CF_INT32:
          // devName in form of trigger:/context/server/device
          trg = devServerName;
          if (trg[0] == '/') return argument_list_error;
          if ((ctx=strchr(trg,':')) == NULL) return argument_list_error;
          *ctx++ = 0;
          if (ctx[0] != '/')
          {
            srv = ctx; ctx = NULL;
          }
          else
          {
            ctx++;
            if ((srv=strchr(ctx,'/')) == NULL) return argument_list_error;
            *srv++ = 0;
          }
          if ((dev=strchr(srv,'/')) != NULL) *dev++ = 0;
          n = dout->dArrayLength;
          n64 = (NAME64 *)dout->data.vptr;
          cc = getArchiveProperties(din->data.lptr[0],trg,ctx,srv,dev,n64,&n);
          dout->dArrayLength = n;
          return cc;
        case CF_NAME16:
        case CF_NAME32:
        case CF_NAME64:
          nlen = fmtsizeof(din->dFormat);
          break;
        default:
          return illegal_format;
      }
      strncpy(devn,din->data.cptr,nlen);
    }
    switch (dout->dFormat)
    {
      case CF_NAME64:
        n64 = (NAME64 *)dout->data.vptr;
        break;
      case CF_NAME32:
      case CF_NAME16:
        n64 = (NAME64 *)qryMemPool;
        break;
      default:
        return illegal_format;
    }
    n = queryDeviceProperties(n64,devServerName,devn,dout->dArrayLength);
    if (n < 0) return -n;
    switch (dout->dFormat)
    {
      case CF_NAME32:
        N1toN2((NAME32 *)dout->data.vptr,n64,dout->dArrayLength,32);
        break;
      case CF_NAME16:
        N1toN2((NAME16 *)dout->data.vptr,n64,dout->dArrayLength,16);
        break;
    }
    dout->dArrayLength = n;
    return 0;
  }

  if (!stricmp(devProperty,"NADEVICES") || !stricmp(devProperty,"NumStoredDevices"))
  {
    int evnnum = 0;
    char context[CONTEXT_NAME_SIZE];
    context[0]  = 0;
    if (strlen(devName) == 0) return illegal_device_number;
    strncpy(devServerName,devName,DEVICE_NAME_SIZE);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength != 1) return illegal_read_write;
      switch (din->dFormat)
      {
        case CF_LONG: 
             evnnum = din->data.lptr[0];
             break;
        case CF_NAME32:
        case CF_NAME16:
             nlen = fmtsizeof(din->dFormat);
             strncpy(context,din->data.cptr,nlen);
             break;
        case CF_NAME32I:
             strncpy(context,((NAME32I *)din->data.vptr)[0].name,32);
             evnnum = ((NAME32I *)din->data.vptr)[0].ival;
             break;
        case CF_NAME16I:
             strncpy(context,((NAME16I *)din->data.vptr)[0].name,16);
             evnnum = ((NAME16I *)din->data.vptr)[0].ival;
             break;
        default: return illegal_format;
      }
    }
    if (dout->dArrayLength != 1) return illegal_read_write;
    switch (dout->dFormat)
    {
      case CF_INTEGER: *dout->data.sptr = queryNumDevices(devServerName,context,evnnum); break;
      case CF_LONG: *dout->data.lptr = queryNumDevices(devServerName,context,evnnum); break;
      default: return illegal_format;
    }
    return 0;
  }

  if (!stricmp(devProperty,"ADEVICES") || !stricmp(devProperty,"StoredDevices"))
  {
    int evnnum = 0;
    char context[16];
    context[0]  = 0;
    if (strlen(devName) == 0) return illegal_device_number;
    strncpy(devServerName,devName,128);
    if (din->dArrayLength > 0)
    {
      if (din->dArrayLength != 1) return illegal_read_write;
      switch (din->dFormat)
      {
        case CF_LONG: 
             evnnum = din->data.lptr[0];
             if (devServerName[0] != '/' && strchr(devServerName,':') != NULL)
             { // devName in form of trigger:/context/server[property]
              trg = devServerName;
              if ((ctx=strchr(trg,':')) == NULL) return argument_list_error;
              *ctx++ = 0;
              if (ctx[0] != '/')
              {
                srv = ctx; ctx = NULL;
              }
              else
              {
                ctx++;
                if ((srv=strchr(ctx,'/')) == NULL) return argument_list_error;
                *srv++ = 0;
              }
              if ((prp=strchr(srv,'[')) != NULL) 
              {
                *prp++ = 0;
                if ((c=strchr(prp,']')) != NULL) *c = 0;
              }
              n = dout->dArrayLength;
              n64 = (NAME64 *)dout->data.vptr;
              cc = getArchiveDevices(evnnum,trg,ctx,srv,prp,n64,&n);
              dout->dArrayLength = n;
              return cc;
             }
             break;
        case CF_NAME32:
        case CF_NAME16:
             nlen = fmtsizeof(din->dFormat);
             strncpy(context,din->data.cptr,nlen);
             break;
        case CF_NAME32I:
             strncpy(context,((NAME32I *)din->data.vptr)[0].name,32);
             evnnum = ((NAME32I *)din->data.vptr)[0].ival;
             break;
        case CF_NAME16I:
             strncpy(context,((NAME16I *)din->data.vptr)[0].name,16);
             evnnum = ((NAME16I *)din->data.vptr)[0].ival;
             break;
        default: return illegal_format;
      }
    }
    switch (dout->dFormat)
    {
      case CF_NAME64:
        n64 = (NAME64 *)dout->data.vptr;
        break;
      case CF_NAME32:
      case CF_NAME16:
        n64 = (NAME64 *)qryMemPool;
        break;
      default:
        return illegal_format;
    }
    n = queryDevices(n64,devServerName,context,evnnum,dout->dArrayLength);
    if (n < 0) return -n;
    switch (dout->dFormat)
    {
      case CF_NAME32:
        N1toN2((NAME32 *)dout->data.vptr,n64,dout->dArrayLength,32);
        break;
      case CF_NAME16:
        N1toN2((NAME16 *)dout->data.vptr,n64,dout->dArrayLength,16);
        break;
    }
    dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"NDEVEVENTS") || !stricmp(devProperty,"NumDeviceEvents"))
  {
    int start = 0, stop = time(NULL);
    if (strlen(devName) == 0) return illegal_device_number;
    strncpy(devServerName,devName,DEVICE_NAME_SIZE); 
    strncpy(device,DEF_DEV_NAME,16); 
    if (din->dArrayLength > 0)
    {
      switch (din->dFormat)
      {
        case CF_LONG: if (din->dArrayLength != 2) return illegal_format;
                      start = din->data.lptr[0];
                      stop  = din->data.lptr[1];
                      break;
        case CF_NAME16I: if (din->dArrayLength != 2) return illegal_format;
                      start = ((NAME16I *)din->data.vptr)[0].ival;
                      stop = ((NAME16I *)din->data.vptr)[1].ival;
                      strncpy(device,((NAME16I *)din->data.vptr)[0].name,16);
                      break;
      }
    }
    nevn = queryNumDeviceArchivedEvents(devServerName,device,start,stop);
    switch (dout->dFormat)
    {
      case CF_SHORT: dout->data.sptr[0] = (UINT16)nevn; break;
      case CF_LONG: dout->data.lptr[0] = nevn; break;
      default: return illegal_format;
    }
    return nevn < 0 ? -nevn : 0;
  }
  if (!stricmp(devProperty,"DEVEVENTS") || !stricmp(devProperty,"DeviceEvents"))
  {
    int stop = time(NULL),start = stop - 86400 * 3; /* defaults: 3 days ago -> now */
    /* devName must be of the form: <trigger>/<context>/<server>/<device>[property] */
    ctx = srv = dev = NULL;
    if ((c=strchr(devName,'/')) != NULL) 
    {
      if ((c=strchr(devName,':')) != NULL) 
      { /* or of the form: <trigger>:/<context>/<server>/<device>[property] */
        *c++ = 0; 
        if (*c != '/') return argument_list_error;
      }
      *c++ = 0; ctx = c;
      if ((c=strchr(ctx,'/')) != NULL)
      {
        *c++ = 0; srv = c;
        if ((c=strchr(srv,'/')) != NULL)
        {
          *c++ = 0; dev = c;
          if ((c=strchr(dev,'[')) != NULL) *c++ = 0;
        }
      }
    }
    if (strlen(devName) == 0) return illegal_device_number;
    if (din->dArrayLength > 0)
    {
      switch (din->dFormat)
      {
        case CF_LONG: 
          if (din->dArrayLength < 2) return dimension_error;
          if (din->dArrayLength > 2) reverseSort = din->data.lptr[2];
          start = din->data.lptr[0];
          stop  = din->data.lptr[1];
          break;
        case CF_NAME16I: 
          if (din->dArrayLength < 2) return dimension_error;
          start = ((NAME16I *)din->data.vptr)[0].ival;
          stop = ((NAME16I *)din->data.vptr)[1].ival;
          strncpy(device,((NAME16I *)din->data.vptr)[0].name,16);
          break;
        default:
          return illegal_format;
      }
    }
    if (dout->dFormat == CF_STRUCT && !strcmp(dout->dTag,"AES") ) 
    {
      n = queryDeviceArchivedEvents(devName,srv,dev,start,stop,(AES *)dout->data.vptr,dout->dArrayLength,reverseSort);
    }
    else if (dout->dFormat == CF_LONG)
    {
      n = queryDeviceArchivedEventList(devName,srv,dev,start,stop,dout->data.ulptr,dout->dArrayLength,reverseSort);
    }
    else
    {
      return illegal_format;
    }
    if (n < 0) return -n;
    if (n < (int)dout->dArrayLength) dout->dArrayLength = (UINT32)n;
    return 0;
  }

  if (!stricmp(devProperty,"REMOVE"))
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dFormat != CF_LONG) return illegal_format;
    switch (din->dArrayLength)
    {
      case 2:
        cc = RemoveArchiveEventsFunction(din->data.lptr[0],din->data.lptr[1],devName);
        if (cc == not_implemented) cc = RemoveArchiveEventFunction(din->data.lptr[0],devName);
        break;
      case 1:
        cc = RemoveArchiveEventsFunction(din->data.lptr[0],din->data.lptr[0],devName);
        if (cc == not_implemented) cc = RemoveArchiveEventFunction(din->data.lptr[0],devName);
        break;
      default:
        return dimension_error;
    }
    return cc;
  }

  if (!stricmp(devProperty,"ANNOTATE"))
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    if (din->dFormat != CF_TEXT) return illegal_format;
    if (din->dArrayLength == 0) return dimension_error;
    if ((cc=parseAnnotationAndEvent(din->data.cptr,&c,&eventTime)) != 0) return cc;
    eventNumber = (UINT32)eventTime;
    return AnnotateArchiveEventFunction(eventNumber,devName,c);
  }

  if (!stricmp(devProperty,"TRC.COMPONENTS.NAM"))
  {
    if (din->dArrayLength > 0 )
    { 
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (!isAdministrator(caller)) return access_denied;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,"TrgView")) return invalid_structure_tag;
      cc = writeViewerFiles((TrgViewFile *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
      if ((cc=readViewerFiles()) != 0) return cc;
      resetPropertyDeviceList(ARCEQM_TAG,"TRC.COMPONENTS");
    }
    if (dout->dArrayLength == 0) return 0;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        nlen = fmtsizeof(dout->dFormat);
        break;
      case CF_STRUCT:
        if (strcmp(dout->dTag,"TrgView")) return invalid_structure_tag;
        nlen = 0;
        break;
      default: return illegal_format;
    }
    if (gNumTrgViewFiles == 0) readViewerFiles();
    for (i=0; i<gNumTrgViewFiles && i<(int)dout->dArrayLength; i++) 
    {
      if (nlen > 0)
      {
        strncpy(dout->data.cptr + i*nlen,trgViewFiles[i].desc,nlen);
      }
      else
      {
        ((TrgViewFile *)dout->data.vptr)[i] = trgViewFiles[i];
      }
    }
    dout->dArrayLength = i;
    return 0;
  }
  if (!stricmp(devProperty,"TRC.COMPONENTS"))
  {
    if ((access & CA_WRITE)) return illegal_read_write;
    if (dout->dArrayLength == 0) return 0;
    if (dout->dFormat != CF_STRUCT) return illegal_format;
    if (strcmp(dout->dTag,"TrgCmp") != 0) return invalid_structure_tag;
    n = GetTriggerComponents(devName,(TrgChannel *)dout->data.vptr,dout->dArrayLength);
    if (n < 0) return -n;
    dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"TRC.CHANNELS.NAM"))
  {
    if (din->dArrayLength > 0 )
    { /* use the WRITE to update the components (filename vs. component name) for a config (devName) */
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (!isAdministrator(caller)) return access_denied;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,"TrgView")) return invalid_structure_tag;
      cc = writeTriggerComponentFile(devName,(TrgViewFile *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
      for (i=0; i<(int)din->dArrayLength; i++) readTriggerComponentFiles(&((TrgViewFile *)din->data.vptr)[i]);
      resetPropertyDeviceList(ARCEQM_TAG,"TRC.CHANNELS");
    }
    if (dout->dArrayLength == 0) return 0;
    switch (dout->dFormat)
    {
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME64:
        nlen = fmtsizeof(dout->dFormat);
        break;
      case CF_STRUCT:
        if (strcmp(dout->dTag,"TrgView")) return invalid_structure_tag;
        nlen = 0;
        break;
      default: return illegal_format;
    }
    if (gNumTrgViewFiles == 0) readViewerFiles();
    for (n=0,i=0; i<gNumTrgViewFiles && n<(int)dout->dArrayLength; i++) 
      for (k=0; k<trgViewFiles[i].nComponents && n<(int)dout->dArrayLength; k++)
      {
        if (strlen(trgViewFiles[i].components[k].desc) == 0) continue;
        if (nlen > 0)
        { /* a name list requested (no duplicates!) */
          for (p=0; p<n; p++)
            if (!strnicmp(dout->data.cptr + p*nlen,trgViewFiles[i].components[k].desc,nlen)) break;
          if (p < n) continue;
          strncpy(dout->data.cptr + n*nlen,trgViewFiles[i].components[k].desc,nlen);
        }
        else
        { /* a list of TrgView structs requested */
          if (stricmp(devName,"ALL") && strnicmp(devName,trgViewFiles[i].desc,64)) 
          { /* config != requested */
            continue;
          }
          for (p=0; p<n; p++) 
            if (!strnicmp(((TrgViewFile *)dout->data.vptr)[p].desc,trgViewFiles[i].components[k].desc,64)) break;
          if (p < n) continue;
          strncpy(((TrgViewFile *)dout->data.vptr)[n].desc,trgViewFiles[i].components[k].desc,64);
          strncpy(((TrgViewFile *)dout->data.vptr)[n].file,trgViewFiles[i].components[k].fileName,64);
        }
        n++;
      }
    dout->dArrayLength = n;
    return 0;
  }
  if (!stricmp(devProperty,"TRC.CHANNELS"))
  {
    if (din->dArrayLength > 0 )
    { 
      if (!(access & CA_WRITE)) return illegal_read_write;
      if (!isAdministrator(caller)) return access_denied;
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strcmp(din->dTag,"TrgChn")) return invalid_structure_tag;
      cc = writeTriggerChannelFile(devName,(TrgChannel *)din->data.vptr,din->dArrayLength);
      if (cc != 0) return cc;
    }
    if (dout->dArrayLength == 0) return 0;
    if (dout->dFormat != CF_STRUCT) return illegal_format;
    if (strcmp(dout->dTag,"TrgChn") != 0) return invalid_structure_tag;
    n = GetTriggerChannels(devName,(TrgChannel *)dout->data.vptr,dout->dArrayLength);
    if (n < 0) return -n;
    dout->dArrayLength = n;
    return 0;
  }
  /* anything left will be a .NAM request of one of the properties */
  if ((c=strrchr(devProperty,'.')) != NULL && !strcmp(c,".NAM"))
  {
    c = "target";
    if (dout->dArrayLength > 1) dout->dArrayLength = 1;
    return PutValuesFromString(dout,c,CF_TEXT,strlen(c),0);
  }

  return illegal_property;
}
PM_EXPORT int initArchiveServer(void)
{
  int i, n;
  NAME32 n32[256];
  char exp[32];
  DTYPE dout, din;

  strncpy(exp,ARCEQM_EXPORT,28);
  if (strlen(thisExtension) > 0) strncat(exp,thisExtension,4);
  RegisterEquipmentModule(exp,ARCEQM_TAG,256,arceqm,arcini,arctsk,100,NULL);

  RegisterProperty(ARCEQM_TAG,"REMOVE",1,CF_LONG,CA_WRITE,"remove event");
  RegisterProperty(ARCEQM_TAG,"REMOVE.NAM",100,CF_NAME64,CA_READ,"'REMOVE' property keywords");
  RegisterProperty(ARCEQM_TAG,"ANNOTATE",1024,CF_TEXT,CA_WRITE,"annotate event");
  RegisterProperty(ARCEQM_TAG,"ANNOTATE.NAM",100,CF_NAME64,CA_READ,"'ANNOTATE' property keywords");

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"AESNOT");
  RegisterPropertyInformation(ARCEQM_TAG,"TRIGGER",NULL,&din,CA_WRITE,AT_SCALAR,0,"receive event trigger",-1,NULL);
  RegisterProperty(ARCEQM_TAG,"TRIGGER.NAM",1,CF_NAME64,CA_READ,"'TRIGGER' property keywords");

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"AESCON");
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"AIB");
  RegisterPropertyInformation(ARCEQM_TAG,"ARCHINFO",&dout,&din,CA_READ|CA_WRITE,AT_SCALAR,0,"archive info query",-1,NULL);
  RegisterPropertyInformation(ARCEQM_TAG,"ArchiveInformation",&dout,&din,CA_READ|CA_WRITE,AT_SCALAR,0,"num properties/devices, event code",-1,NULL);

  RegisterProperty(ARCEQM_TAG,"ARCHSTATE",1,CF_LONG,CA_WRITE|CA_READ,"read/set archive trigger state");
  RegisterProperty(ARCEQM_TAG,"ArchiveTriggerStatus",1,CF_LONG,CA_WRITE|CA_READ,"read/set archive trigger status");
  RegisterProperty(ARCEQM_TAG,"ARCHWINDOW",1,CF_LONG,CA_WRITE|CA_READ,"read/set archive trigger window");
  RegisterProperty(ARCEQM_TAG,"ArchiveTriggerWindow",1,CF_LONG,CA_WRITE|CA_READ,"read/set archive trigger window");

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"AESCON");
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"PIB");
  RegisterPropertyInformation(ARCEQM_TAG,"PROPINFO",&dout,&din,CA_READ,AT_SCALAR,0,"property info query",-1,NULL);
  RegisterProperty(ARCEQM_TAG,"PROPINFO.NAM",1,CF_NAME64,CA_READ,"'PROPINFO' property keywords");

  din.dFormat = CF_LONG;
  din.dArrayLength = 1;
  din.dTag[0] = 0;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"PRPQSr4");
  RegisterPropertyInformation(ARCEQM_TAG,"PropertyInformation",&dout,&din,CA_READ,AT_SCALAR,0,"property information header",-1,NULL);
  RegisterProperty(ARCEQM_TAG,"PropertyInformation.NAM",1,CF_NAME64,CA_READ,"'PropertyInformation' property keywords");

  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"ADRB");
  dout.dFormat = CF_BYTE;
  dout.dArrayLength = 4096;
  dout.dTag[0] = 0;
  RegisterPropertyInformation(ARCEQM_TAG,"ARCHDATA",&dout,&din,CA_READ,AT_SCALAR,0,"get archived data",-1,NULL);
  RegisterPropertyInformation(ARCEQM_TAG,"StoredData",&dout,&din,CA_READ,AT_SCALAR,0,"stored data for given archive data request",-1,NULL);

  din.dFormat = CF_INT32;
  din.dArrayLength = 1;
  din.dTag[0] = 0;
  RegisterPropertyInformation(ARCEQM_TAG,"CloseArchive",NULL,&din,CA_READ,AT_SCALAR,0,"closed the targeted archive session",-1,NULL);

  RegisterProperty(ARCEQM_TAG,"ARCHDATA.NAM",100,CF_NAME64,CA_READ,"'ARCHDATA' property keywords");
  RegisterProperty(ARCEQM_TAG,"StoredData.NAM",100,CF_NAME64,CA_READ,"'StoredData' property keywords");
  RegisterProperty(ARCEQM_TAG,"ARCHHEADER.NAM",100,CF_NAME64,CA_READ,"'ARCHHEADER' property keywords");
  RegisterProperty(ARCEQM_TAG,"DataHeader.NAM",100,CF_NAME64,CA_READ,"'DataHeader' property keywords");
  RegisterProperty(ARCEQM_TAG,"CloseArchive.NAM",100,CF_NAME64,CA_READ,"'CloseArchive' property keywords");
  
  din.dFormat = CF_STRUCT;
  din.dArrayLength = 1;
  strcpy(din.dTag,"ADRB");
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"DHS");
  RegisterPropertyInformation(ARCEQM_TAG,"ARCHHEADER",&dout,&din,CA_READ,AT_SCALAR,0,"get archive data header",-1,NULL);
  RegisterPropertyInformation(ARCEQM_TAG,"DataHeader",&dout,&din,CA_READ,AT_SCALAR,0,"data header for given archive data request",-1,NULL);

  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 1;
  strcpy(dout.dTag,"EVNDHDR");
  RegisterPropertyInformation(ARCEQM_TAG,"DataHeader",&dout,&din,CA_READ,AT_SCALAR,0,"data header for given archive data request",-1,NULL);

  RegisterProperty(ARCEQM_TAG,"NEvents",1,CF_LONG,CA_READ,"number of events in a time interval for given context");
  RegisterProperty(ARCEQM_TAG,"NEvents.NAM",100,CF_NAME64,CA_READ,"archive triggers applicable for property Events");
  RegisterProperty(ARCEQM_TAG,"NumEvents",1,CF_LONG,CA_READ,"number of events in a time interval for given context");
  RegisterProperty(ARCEQM_TAG,"NumEvents.NAM",100,CF_NAME64,CA_READ,"archive triggers applicable for property Events");
  RegisterProperty(ARCEQM_TAG,"Events",1000,CF_LONG,CA_READ,"events in a time interval for given context");
  RegisterProperty(ARCEQM_TAG,"Events.NAM",100,CF_NAME64,CA_READ,"archive triggers applicable for property Events");
  RegisterProperty(ARCEQM_TAG,"NDSERVERS",1,CF_LONG,CA_READ,"number of device servers");
  RegisterProperty(ARCEQM_TAG,"NumStoredServers",1,CF_LONG,CA_READ,"number of device servers");
  RegisterProperty(ARCEQM_TAG,"DSERVERS",100,CF_NAME32,CA_READ,"device server names");
  RegisterProperty(ARCEQM_TAG,"StoredServers",100,CF_NAME32,CA_READ,"stored device server names for event context given");
  RegisterProperty(ARCEQM_TAG,"StoredServers.NAM",100,CF_NAME64,CA_READ,"event context names");
  RegisterProperty(ARCEQM_TAG,"NADEVICES",1,CF_LONG,CA_READ,"number of device servers");
  RegisterProperty(ARCEQM_TAG,"NumStoredDevices",1,CF_LONG,CA_READ,"number of device servers");
  RegisterProperty(ARCEQM_TAG,"ADEVICES",1024,CF_NAME64,CA_READ,"device list");
  RegisterProperty(ARCEQM_TAG,"StoredDevices",1024,CF_NAME64,CA_READ,"stored device list for event and server given");
  RegisterProperty(ARCEQM_TAG,"NAPROPS",1,CF_LONG,CA_READ,"number of device properties");
  RegisterProperty(ARCEQM_TAG,"NumStoredProperties",1,CF_LONG,CA_READ,"number of device properties");
  RegisterProperty(ARCEQM_TAG,"APROPS",100,CF_NAME64,CA_READ,"property list");
  RegisterProperty(ARCEQM_TAG,"StoredProperties",100,CF_NAME64,CA_READ,"stored property list for event and server given");
  RegisterProperty(ARCEQM_TAG,"NDEVEVENTS",1,CF_LONG,CA_READ,"number of device events");
  RegisterProperty(ARCEQM_TAG,"NumDeviceEvents",1,CF_LONG,CA_READ,"number of device events");
  RegisterProperty(ARCEQM_TAG,"DEVEVENTS",100,CF_LONG,CA_READ,"device event list");
  RegisterProperty(ARCEQM_TAG,"DeviceEvents",100,CF_LONG,CA_READ,"device specific event list");
  RegisterProperty(ARCEQM_TAG,"StoredContexts",1024,CF_NAME64,CA_READ,"stored archive context list for server and property given");
  RegisterProperty(ARCEQM_TAG,"Triggers",1024,CF_NAME64,CA_READ,"trigger list for server and property given");

  memset(&din,0,sizeof(DTYPE));
  din.dFormat = CF_NULL;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 64;
  strcpy(dout.dTag,"TrgCmp");
  RegisterPropertyInformation(ARCEQM_TAG,"TRC.COMPONENTS",&dout,&din,CA_READ,AT_CHANNEL,0,"Transient Recorder viewer components",-1,NULL);

  RegisterProperty(ARCEQM_TAG,"TRC.COMPONENTS.NAM",64,CF_NAME64,CA_READ,"component name list");
  din.dArrayLength = 64;
  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"TrgView");
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 64;
  strcpy(dout.dTag,"TrgView");
  RegisterPropertyInformation(ARCEQM_TAG,"TRC.COMPONENTS.NAM",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,0,"Transient Recorder viewer components",-1,NULL);

  strcpy(dout.dTag,"TrgChn");
  RegisterPropertyInformation(ARCEQM_TAG,"TRC.CHANNELS",&dout,&din,CA_READ,AT_SCALAR,0,"Transient Recorder viewer channels",-1,NULL);

  RegisterProperty(ARCEQM_TAG,"TRC.CHANNELS.NAM",64,CF_NAME64,CA_READ,"channel name list");
  din.dArrayLength = 64;
  din.dFormat = CF_STRUCT;
  strcpy(din.dTag,"TrgView");
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = 64;
  strcpy(dout.dTag,"TrgView");
  RegisterPropertyInformation(ARCEQM_TAG,"TRC.CHANNELS.NAM",&dout,&din,CA_READ|CA_WRITE,AT_CHANNEL,0,"channel name list",-1,NULL);

  n = GetLocalDeviceServers(NULL,n32,0,time(NULL),-1,256);
  for (i=0; i<n; i++)
  {
    RegisterDeviceName(ARCEQM_TAG,n32[i].name,i);
  }
  if (n > 0) AssignNumDevices(ARCEQM_TAG,n);

  if (useConnectedLinks)
  {
    feclog("use connected transport communication for large payloads");
  }

  return 0;
}
PM_EXPORT int initArchiveClient(void)
{
  arcini();
  return 0;
}
